#ifndef _ADT_H
#define _ADT_H

/*{{{  includes */

#include <stdlib.h>
#include <string.h>
#include <aterm1.h>
#include "ADT_dict.h"

/*}}}  */

/*{{{  typedefs */

typedef struct _ADT_Entries *ADT_Entries;
typedef struct _ADT_Entry *ADT_Entry;
typedef struct _ADT_Separators *ADT_Separators;
typedef struct _ADT_Separator *ADT_Separator;
typedef struct _ADT_Modules *ADT_Modules;
typedef struct _ADT_Module *ADT_Module;
typedef struct _ADT_Imports *ADT_Imports;
typedef struct _ADT_Type *ADT_Type;
typedef struct _ADT_Sorts *ADT_Sorts;
typedef struct _ADT_ModuleName *ADT_ModuleName;

/*}}}  */

void ADT_initADTApi(void);

/*{{{  (un)protect functions */

void ADT_protectEntries(ADT_Entries *arg);
void ADT_unprotectEntries(ADT_Entries *arg);
void ADT_protectEntry(ADT_Entry *arg);
void ADT_unprotectEntry(ADT_Entry *arg);
void ADT_protectSeparators(ADT_Separators *arg);
void ADT_unprotectSeparators(ADT_Separators *arg);
void ADT_protectSeparator(ADT_Separator *arg);
void ADT_unprotectSeparator(ADT_Separator *arg);
void ADT_protectModules(ADT_Modules *arg);
void ADT_unprotectModules(ADT_Modules *arg);
void ADT_protectModule(ADT_Module *arg);
void ADT_unprotectModule(ADT_Module *arg);
void ADT_protectImports(ADT_Imports *arg);
void ADT_unprotectImports(ADT_Imports *arg);
void ADT_protectType(ADT_Type *arg);
void ADT_unprotectType(ADT_Type *arg);
void ADT_protectSorts(ADT_Sorts *arg);
void ADT_unprotectSorts(ADT_Sorts *arg);
void ADT_protectModuleName(ADT_ModuleName *arg);
void ADT_unprotectModuleName(ADT_ModuleName *arg);

/*}}}  */
/*{{{  term conversion functions */

ADT_Entries ADT_EntriesFromTerm(ATerm t);
ATerm ADT_EntriesToTerm(ADT_Entries arg);
ADT_Entry ADT_EntryFromTerm(ATerm t);
ATerm ADT_EntryToTerm(ADT_Entry arg);
ADT_Separators ADT_SeparatorsFromTerm(ATerm t);
ATerm ADT_SeparatorsToTerm(ADT_Separators arg);
ADT_Separator ADT_SeparatorFromTerm(ATerm t);
ATerm ADT_SeparatorToTerm(ADT_Separator arg);
ADT_Modules ADT_ModulesFromTerm(ATerm t);
ATerm ADT_ModulesToTerm(ADT_Modules arg);
ADT_Module ADT_ModuleFromTerm(ATerm t);
ATerm ADT_ModuleToTerm(ADT_Module arg);
ADT_Imports ADT_ImportsFromTerm(ATerm t);
ATerm ADT_ImportsToTerm(ADT_Imports arg);
ADT_Type ADT_TypeFromTerm(ATerm t);
ATerm ADT_TypeToTerm(ADT_Type arg);
ADT_Sorts ADT_SortsFromTerm(ATerm t);
ATerm ADT_SortsToTerm(ADT_Sorts arg);
ADT_ModuleName ADT_ModuleNameFromTerm(ATerm t);
ATerm ADT_ModuleNameToTerm(ADT_ModuleName arg);

/*}}}  */
/*{{{  list functions */

int ADT_getEntriesLength (ADT_Entries arg);
ADT_Entries ADT_reverseEntries(ADT_Entries arg);
ADT_Entries ADT_appendEntries(ADT_Entries arg, ADT_Entry elem);
ADT_Entries ADT_concatEntries(ADT_Entries arg0, ADT_Entries arg1);
ADT_Entries ADT_sliceEntries(ADT_Entries arg, int start, int end);
ADT_Entry ADT_getEntriesEntryAt(ADT_Entries arg, int index);
ADT_Entries ADT_replaceEntriesEntryAt(ADT_Entries arg, ADT_Entry elem, int index);
ADT_Entries ADT_makeEntries2(ADT_Entry elem1, ADT_Entry elem2);
ADT_Entries ADT_makeEntries3(ADT_Entry elem1, ADT_Entry elem2, ADT_Entry elem3);
ADT_Entries ADT_makeEntries4(ADT_Entry elem1, ADT_Entry elem2, ADT_Entry elem3, ADT_Entry elem4);
ADT_Entries ADT_makeEntries5(ADT_Entry elem1, ADT_Entry elem2, ADT_Entry elem3, ADT_Entry elem4, ADT_Entry elem5);
ADT_Entries ADT_makeEntries6(ADT_Entry elem1, ADT_Entry elem2, ADT_Entry elem3, ADT_Entry elem4, ADT_Entry elem5, ADT_Entry elem6);
int ADT_getSeparatorsLength (ADT_Separators arg);
ADT_Separators ADT_reverseSeparators(ADT_Separators arg);
ADT_Separators ADT_appendSeparators(ADT_Separators arg, ADT_Separator elem);
ADT_Separators ADT_concatSeparators(ADT_Separators arg0, ADT_Separators arg1);
ADT_Separators ADT_sliceSeparators(ADT_Separators arg, int start, int end);
ADT_Separator ADT_getSeparatorsSeparatorAt(ADT_Separators arg, int index);
ADT_Separators ADT_replaceSeparatorsSeparatorAt(ADT_Separators arg, ADT_Separator elem, int index);
ADT_Separators ADT_makeSeparators2(ADT_Separator elem1, ADT_Separator elem2);
ADT_Separators ADT_makeSeparators3(ADT_Separator elem1, ADT_Separator elem2, ADT_Separator elem3);
ADT_Separators ADT_makeSeparators4(ADT_Separator elem1, ADT_Separator elem2, ADT_Separator elem3, ADT_Separator elem4);
ADT_Separators ADT_makeSeparators5(ADT_Separator elem1, ADT_Separator elem2, ADT_Separator elem3, ADT_Separator elem4, ADT_Separator elem5);
ADT_Separators ADT_makeSeparators6(ADT_Separator elem1, ADT_Separator elem2, ADT_Separator elem3, ADT_Separator elem4, ADT_Separator elem5, ADT_Separator elem6);
int ADT_getModulesLength (ADT_Modules arg);
ADT_Modules ADT_reverseModules(ADT_Modules arg);
ADT_Modules ADT_appendModules(ADT_Modules arg, ADT_Module elem);
ADT_Modules ADT_concatModules(ADT_Modules arg0, ADT_Modules arg1);
ADT_Modules ADT_sliceModules(ADT_Modules arg, int start, int end);
ADT_Module ADT_getModulesModuleAt(ADT_Modules arg, int index);
ADT_Modules ADT_replaceModulesModuleAt(ADT_Modules arg, ADT_Module elem, int index);
ADT_Modules ADT_makeModules2(ADT_Module elem1, ADT_Module elem2);
ADT_Modules ADT_makeModules3(ADT_Module elem1, ADT_Module elem2, ADT_Module elem3);
ADT_Modules ADT_makeModules4(ADT_Module elem1, ADT_Module elem2, ADT_Module elem3, ADT_Module elem4);
ADT_Modules ADT_makeModules5(ADT_Module elem1, ADT_Module elem2, ADT_Module elem3, ADT_Module elem4, ADT_Module elem5);
ADT_Modules ADT_makeModules6(ADT_Module elem1, ADT_Module elem2, ADT_Module elem3, ADT_Module elem4, ADT_Module elem5, ADT_Module elem6);
int ADT_getImportsLength (ADT_Imports arg);
ADT_Imports ADT_reverseImports(ADT_Imports arg);
ADT_Imports ADT_appendImports(ADT_Imports arg, ADT_ModuleName elem);
ADT_Imports ADT_concatImports(ADT_Imports arg0, ADT_Imports arg1);
ADT_Imports ADT_sliceImports(ADT_Imports arg, int start, int end);
ADT_ModuleName ADT_getImportsModuleNameAt(ADT_Imports arg, int index);
ADT_Imports ADT_replaceImportsModuleNameAt(ADT_Imports arg, ADT_ModuleName elem, int index);
ADT_Imports ADT_makeImports2(ADT_ModuleName elem1, ADT_ModuleName elem2);
ADT_Imports ADT_makeImports3(ADT_ModuleName elem1, ADT_ModuleName elem2, ADT_ModuleName elem3);
ADT_Imports ADT_makeImports4(ADT_ModuleName elem1, ADT_ModuleName elem2, ADT_ModuleName elem3, ADT_ModuleName elem4);
ADT_Imports ADT_makeImports5(ADT_ModuleName elem1, ADT_ModuleName elem2, ADT_ModuleName elem3, ADT_ModuleName elem4, ADT_ModuleName elem5);
ADT_Imports ADT_makeImports6(ADT_ModuleName elem1, ADT_ModuleName elem2, ADT_ModuleName elem3, ADT_ModuleName elem4, ADT_ModuleName elem5, ADT_ModuleName elem6);
int ADT_getSortsLength (ADT_Sorts arg);
ADT_Sorts ADT_reverseSorts(ADT_Sorts arg);
ADT_Sorts ADT_appendSorts(ADT_Sorts arg, ADT_Type elem);
ADT_Sorts ADT_concatSorts(ADT_Sorts arg0, ADT_Sorts arg1);
ADT_Sorts ADT_sliceSorts(ADT_Sorts arg, int start, int end);
ADT_Type ADT_getSortsTypeAt(ADT_Sorts arg, int index);
ADT_Sorts ADT_replaceSortsTypeAt(ADT_Sorts arg, ADT_Type elem, int index);
ADT_Sorts ADT_makeSorts2(ADT_Type elem1, ADT_Type elem2);
ADT_Sorts ADT_makeSorts3(ADT_Type elem1, ADT_Type elem2, ADT_Type elem3);
ADT_Sorts ADT_makeSorts4(ADT_Type elem1, ADT_Type elem2, ADT_Type elem3, ADT_Type elem4);
ADT_Sorts ADT_makeSorts5(ADT_Type elem1, ADT_Type elem2, ADT_Type elem3, ADT_Type elem4, ADT_Type elem5);
ADT_Sorts ADT_makeSorts6(ADT_Type elem1, ADT_Type elem2, ADT_Type elem3, ADT_Type elem4, ADT_Type elem5, ADT_Type elem6);

/*}}}  */
/*{{{  constructors */

ADT_Entries ADT_makeEntriesEmpty(void);
ADT_Entries ADT_makeEntriesSingle(ADT_Entry head);
ADT_Entries ADT_makeEntriesMany(ADT_Entry head, ADT_Entries tail);
ADT_Entry ADT_makeEntryConstructor(ATerm sort, ATerm alternative, ATerm termPattern);
ADT_Entry ADT_makeEntryList(ATerm sort, ATerm elemSort);
ADT_Entry ADT_makeEntryNamedList(ATerm opname, ATerm sort, ATerm elemSort);
ADT_Entry ADT_makeEntrySeparatedList(ATerm sort, ATerm elemSort, ADT_Separators separators);
ADT_Separators ADT_makeSeparatorsEmpty(void);
ADT_Separators ADT_makeSeparatorsSingle(ADT_Separator head);
ADT_Separators ADT_makeSeparatorsMany(ADT_Separator head, ADT_Separators tail);
ADT_Separator ADT_makeSeparatorDefault(ATerm termPattern);
ADT_Modules ADT_makeModulesEmpty(void);
ADT_Modules ADT_makeModulesSingle(ADT_Module head);
ADT_Modules ADT_makeModulesMany(ADT_Module head, ADT_Modules tail);
ADT_Module ADT_makeModuleModulentry(ADT_ModuleName modulename, ADT_Imports imports, ADT_Sorts sorts, ADT_Entries entries);
ADT_Imports ADT_makeImportsEmpty(void);
ADT_Imports ADT_makeImportsSingle(ADT_ModuleName head);
ADT_Imports ADT_makeImportsMany(ADT_ModuleName head, ADT_Imports tail);
ADT_Type ADT_makeTypeType(const char* name);
ADT_Sorts ADT_makeSortsEmpty(void);
ADT_Sorts ADT_makeSortsSingle(ADT_Type head);
ADT_Sorts ADT_makeSortsMany(ADT_Type head, ADT_Sorts tail);
ADT_ModuleName ADT_makeModuleNameName(const char* name);

/*}}}  */
/*{{{  equality functions */

ATbool ADT_isEqualEntries(ADT_Entries arg0, ADT_Entries arg1);
ATbool ADT_isEqualEntry(ADT_Entry arg0, ADT_Entry arg1);
ATbool ADT_isEqualSeparators(ADT_Separators arg0, ADT_Separators arg1);
ATbool ADT_isEqualSeparator(ADT_Separator arg0, ADT_Separator arg1);
ATbool ADT_isEqualModules(ADT_Modules arg0, ADT_Modules arg1);
ATbool ADT_isEqualModule(ADT_Module arg0, ADT_Module arg1);
ATbool ADT_isEqualImports(ADT_Imports arg0, ADT_Imports arg1);
ATbool ADT_isEqualType(ADT_Type arg0, ADT_Type arg1);
ATbool ADT_isEqualSorts(ADT_Sorts arg0, ADT_Sorts arg1);
ATbool ADT_isEqualModuleName(ADT_ModuleName arg0, ADT_ModuleName arg1);

/*}}}  */
/*{{{  ADT_Entries accessors */

ATbool ADT_isValidEntries(ADT_Entries arg);
inline ATbool ADT_isEntriesEmpty(ADT_Entries arg);
inline ATbool ADT_isEntriesSingle(ADT_Entries arg);
inline ATbool ADT_isEntriesMany(ADT_Entries arg);
ATbool ADT_hasEntriesHead(ADT_Entries arg);
ATbool ADT_hasEntriesTail(ADT_Entries arg);
ADT_Entry ADT_getEntriesHead(ADT_Entries arg);
ADT_Entries ADT_getEntriesTail(ADT_Entries arg);
ADT_Entries ADT_setEntriesHead(ADT_Entries arg, ADT_Entry head);
ADT_Entries ADT_setEntriesTail(ADT_Entries arg, ADT_Entries tail);

/*}}}  */
/*{{{  ADT_Entry accessors */

ATbool ADT_isValidEntry(ADT_Entry arg);
inline ATbool ADT_isEntryConstructor(ADT_Entry arg);
inline ATbool ADT_isEntryList(ADT_Entry arg);
inline ATbool ADT_isEntryNamedList(ADT_Entry arg);
inline ATbool ADT_isEntrySeparatedList(ADT_Entry arg);
ATbool ADT_hasEntrySort(ADT_Entry arg);
ATbool ADT_hasEntryAlternative(ADT_Entry arg);
ATbool ADT_hasEntryTermPattern(ADT_Entry arg);
ATbool ADT_hasEntryElemSort(ADT_Entry arg);
ATbool ADT_hasEntryOpname(ADT_Entry arg);
ATbool ADT_hasEntrySeparators(ADT_Entry arg);
ATerm ADT_getEntrySort(ADT_Entry arg);
ATerm ADT_getEntryAlternative(ADT_Entry arg);
ATerm ADT_getEntryTermPattern(ADT_Entry arg);
ATerm ADT_getEntryElemSort(ADT_Entry arg);
ATerm ADT_getEntryOpname(ADT_Entry arg);
ADT_Separators ADT_getEntrySeparators(ADT_Entry arg);
ADT_Entry ADT_setEntrySort(ADT_Entry arg, ATerm sort);
ADT_Entry ADT_setEntryAlternative(ADT_Entry arg, ATerm alternative);
ADT_Entry ADT_setEntryTermPattern(ADT_Entry arg, ATerm termPattern);
ADT_Entry ADT_setEntryElemSort(ADT_Entry arg, ATerm elemSort);
ADT_Entry ADT_setEntryOpname(ADT_Entry arg, ATerm opname);
ADT_Entry ADT_setEntrySeparators(ADT_Entry arg, ADT_Separators separators);

/*}}}  */
/*{{{  ADT_Separators accessors */

ATbool ADT_isValidSeparators(ADT_Separators arg);
inline ATbool ADT_isSeparatorsEmpty(ADT_Separators arg);
inline ATbool ADT_isSeparatorsSingle(ADT_Separators arg);
inline ATbool ADT_isSeparatorsMany(ADT_Separators arg);
ATbool ADT_hasSeparatorsHead(ADT_Separators arg);
ATbool ADT_hasSeparatorsTail(ADT_Separators arg);
ADT_Separator ADT_getSeparatorsHead(ADT_Separators arg);
ADT_Separators ADT_getSeparatorsTail(ADT_Separators arg);
ADT_Separators ADT_setSeparatorsHead(ADT_Separators arg, ADT_Separator head);
ADT_Separators ADT_setSeparatorsTail(ADT_Separators arg, ADT_Separators tail);

/*}}}  */
/*{{{  ADT_Separator accessors */

ATbool ADT_isValidSeparator(ADT_Separator arg);
inline ATbool ADT_isSeparatorDefault(ADT_Separator arg);
ATbool ADT_hasSeparatorTermPattern(ADT_Separator arg);
ATerm ADT_getSeparatorTermPattern(ADT_Separator arg);
ADT_Separator ADT_setSeparatorTermPattern(ADT_Separator arg, ATerm termPattern);

/*}}}  */
/*{{{  ADT_Modules accessors */

ATbool ADT_isValidModules(ADT_Modules arg);
inline ATbool ADT_isModulesEmpty(ADT_Modules arg);
inline ATbool ADT_isModulesSingle(ADT_Modules arg);
inline ATbool ADT_isModulesMany(ADT_Modules arg);
ATbool ADT_hasModulesHead(ADT_Modules arg);
ATbool ADT_hasModulesTail(ADT_Modules arg);
ADT_Module ADT_getModulesHead(ADT_Modules arg);
ADT_Modules ADT_getModulesTail(ADT_Modules arg);
ADT_Modules ADT_setModulesHead(ADT_Modules arg, ADT_Module head);
ADT_Modules ADT_setModulesTail(ADT_Modules arg, ADT_Modules tail);

/*}}}  */
/*{{{  ADT_Module accessors */

ATbool ADT_isValidModule(ADT_Module arg);
inline ATbool ADT_isModuleModulentry(ADT_Module arg);
ATbool ADT_hasModuleModulename(ADT_Module arg);
ATbool ADT_hasModuleImports(ADT_Module arg);
ATbool ADT_hasModuleSorts(ADT_Module arg);
ATbool ADT_hasModuleEntries(ADT_Module arg);
ADT_ModuleName ADT_getModuleModulename(ADT_Module arg);
ADT_Imports ADT_getModuleImports(ADT_Module arg);
ADT_Sorts ADT_getModuleSorts(ADT_Module arg);
ADT_Entries ADT_getModuleEntries(ADT_Module arg);
ADT_Module ADT_setModuleModulename(ADT_Module arg, ADT_ModuleName modulename);
ADT_Module ADT_setModuleImports(ADT_Module arg, ADT_Imports imports);
ADT_Module ADT_setModuleSorts(ADT_Module arg, ADT_Sorts sorts);
ADT_Module ADT_setModuleEntries(ADT_Module arg, ADT_Entries entries);

/*}}}  */
/*{{{  ADT_Imports accessors */

ATbool ADT_isValidImports(ADT_Imports arg);
inline ATbool ADT_isImportsEmpty(ADT_Imports arg);
inline ATbool ADT_isImportsSingle(ADT_Imports arg);
inline ATbool ADT_isImportsMany(ADT_Imports arg);
ATbool ADT_hasImportsHead(ADT_Imports arg);
ATbool ADT_hasImportsTail(ADT_Imports arg);
ADT_ModuleName ADT_getImportsHead(ADT_Imports arg);
ADT_Imports ADT_getImportsTail(ADT_Imports arg);
ADT_Imports ADT_setImportsHead(ADT_Imports arg, ADT_ModuleName head);
ADT_Imports ADT_setImportsTail(ADT_Imports arg, ADT_Imports tail);

/*}}}  */
/*{{{  ADT_Type accessors */

ATbool ADT_isValidType(ADT_Type arg);
inline ATbool ADT_isTypeType(ADT_Type arg);
ATbool ADT_hasTypeName(ADT_Type arg);
char* ADT_getTypeName(ADT_Type arg);
ADT_Type ADT_setTypeName(ADT_Type arg, const char* name);

/*}}}  */
/*{{{  ADT_Sorts accessors */

ATbool ADT_isValidSorts(ADT_Sorts arg);
inline ATbool ADT_isSortsEmpty(ADT_Sorts arg);
inline ATbool ADT_isSortsSingle(ADT_Sorts arg);
inline ATbool ADT_isSortsMany(ADT_Sorts arg);
ATbool ADT_hasSortsHead(ADT_Sorts arg);
ATbool ADT_hasSortsTail(ADT_Sorts arg);
ADT_Type ADT_getSortsHead(ADT_Sorts arg);
ADT_Sorts ADT_getSortsTail(ADT_Sorts arg);
ADT_Sorts ADT_setSortsHead(ADT_Sorts arg, ADT_Type head);
ADT_Sorts ADT_setSortsTail(ADT_Sorts arg, ADT_Sorts tail);

/*}}}  */
/*{{{  ADT_ModuleName accessors */

ATbool ADT_isValidModuleName(ADT_ModuleName arg);
inline ATbool ADT_isModuleNameName(ADT_ModuleName arg);
ATbool ADT_hasModuleNameName(ADT_ModuleName arg);
char* ADT_getModuleNameName(ADT_ModuleName arg);
ADT_ModuleName ADT_setModuleNameName(ADT_ModuleName arg, const char* name);

/*}}}  */
/*{{{  sort visitors */

ADT_Entries ADT_visitEntries(ADT_Entries arg, ADT_Entry (*acceptHead)(ADT_Entry));
ADT_Entry ADT_visitEntry(ADT_Entry arg, ATerm (*acceptSort)(ATerm), ATerm (*acceptAlternative)(ATerm), ATerm (*acceptTermPattern)(ATerm), ATerm (*acceptElemSort)(ATerm), ATerm (*acceptOpname)(ATerm), ADT_Separators (*acceptSeparators)(ADT_Separators));
ADT_Separators ADT_visitSeparators(ADT_Separators arg, ADT_Separator (*acceptHead)(ADT_Separator));
ADT_Separator ADT_visitSeparator(ADT_Separator arg, ATerm (*acceptTermPattern)(ATerm));
ADT_Modules ADT_visitModules(ADT_Modules arg, ADT_Module (*acceptHead)(ADT_Module));
ADT_Module ADT_visitModule(ADT_Module arg, ADT_ModuleName (*acceptModulename)(ADT_ModuleName), ADT_Imports (*acceptImports)(ADT_Imports), ADT_Sorts (*acceptSorts)(ADT_Sorts), ADT_Entries (*acceptEntries)(ADT_Entries));
ADT_Imports ADT_visitImports(ADT_Imports arg, ADT_ModuleName (*acceptHead)(ADT_ModuleName));
ADT_Type ADT_visitType(ADT_Type arg, char* (*acceptName)(char*));
ADT_Sorts ADT_visitSorts(ADT_Sorts arg, ADT_Type (*acceptHead)(ADT_Type));
ADT_ModuleName ADT_visitModuleName(ADT_ModuleName arg, char* (*acceptName)(char*));

/*}}}  */

#endif /* _ADT_H */
